#include <header.h>

//thread log
struct log_t	tlog;

void tlog_record( short priority, char *msg, struct vhost_t *vhost, struct http_request_header_t *request_header, struct http_response_header_t *response_header )
{
	struct log_record_t		*new_record;
	extern time_t			now;


	new_record = malloc( sizeof(struct log_record_t) );
	new_record->next = NULL;
	new_record->request_referer = malloc( LOG_URL_SIZE * 2 );
	new_record->request_url = &(new_record->request_referer[LOG_URL_SIZE]);
	new_record->time = now;
	new_record->priority = priority;
	new_record->msg = msg;
	new_record->client_addr.s_addr = request_header->cgi_env.addr->sin_addr.s_addr;
	if ( request_header->referer != NULL )
		new_record->request_referer = strncpy( new_record->request_referer, request_header->referer, LOG_URL_SIZE );
	if ( request_header->encoded_url != NULL )
		new_record->request_url = strncpy( new_record->request_url, request_header->encoded_url, LOG_URL_SIZE );
	new_record->request_method = request_header->method;
	new_record->response_status = response_header->status;
	new_record->response_content_length = response_header->content_length;

	if ( vhost != NULL )
		tlog_record_append( vhost->logfile, new_record );
	else
	{
		new_record->request_vhost = NULL;
		tlog_record_append( &(tlog.logfile[0]), new_record );
	}

	return;
}

void tlog_record_append( struct logfile_t *logfile, struct log_record_t *record )
{
	pthread_mutex_lock( &logfile->mutex );
	if ( logfile->log_list_start == NULL )
		logfile->log_list_start = logfile->log_list_end = record;
	else
	{
		logfile->log_list_end->next = record;
		logfile->log_list_end = record;
	}
	++logfile->log_list_size;
	pthread_mutex_unlock( &logfile->mutex );
	return;
}


void tlog_init()
{
	extern time_t		now;
	struct tm		now_tm;
	int			i;

	pthread_mutex_init( &tlog.mutex, NULL );

	localtime_r( &now, &now_tm );

	fprintf(stderr, "*** tlog initializing.\n");
	while ( tlog_open_logfiles( &now_tm ) == FALSE )
	{
		fprintf(stderr, "*** tlog file open failed. retry.\n");
		sleep(1);
	}

	for ( i=tlog.logfile_len; i--; )
	{
		pthread_mutex_init( &(tlog.logfile[i].mutex), NULL );
		if ( tlog.logfile[i].vhost != NULL )
			tlog.logfile[i].vhost->logfile = &(tlog.logfile[i]);
	}

	return;
}


void *tlog_proc(void *data)
{
	extern time_t		now;
	struct tm		now_tm;
	int			i;

	localtime_r( &now, &now_tm );

	fprintf(stderr, "*** tlog thread start.\n");

	while ( 1 )
	{
		localtime_r( &now, &now_tm );
		sleep(1);
		//usleep(100000);
		for ( i=tlog.logfile_len; i--; )
		{
			if ( tlog.logfile[i].log_list_size > 0 )
				tlog_write_logfile( &(tlog.logfile[i]), &now_tm );
		}
	}

	pthread_exit(0);
}

void tlog_write_logfile( struct logfile_t *logfile, struct tm *now )
{
	struct log_record_t	*record;
	char			log_msg[LOG_RECORD_LEN];
	char			asctime_buf[26];
	char			*method;
	struct tm		tm;
	size_t			log_len;

	pthread_mutex_lock( &logfile->mutex );
	if ( logfile->log_list_start == NULL )
	{
		pthread_mutex_unlock( &logfile->mutex );
		return;
	}
	pthread_mutex_unlock( &logfile->mutex );

	while ( logfile->log_list_size > 0 )
	{
		record = logfile->log_list_start;
		pthread_mutex_lock( &logfile->mutex );
		if ( record->next == NULL )
		{
			logfile->log_list_start = logfile->log_list_end = NULL;
			logfile->log_list_size = 0;
		}
		else
		{
			logfile->log_list_start = record->next;
			--logfile->log_list_size;
		}
		pthread_mutex_unlock( &logfile->mutex );

		localtime_r( &record->time, &tm );
		if ( tlog.rotate == TRUE && (logfile->time + tlog.rotate_cycle) < record->time )
		{
			tlog_close_logfile( logfile );
			tlog_open_logfile( logfile, &tm );
		}
		asctime_r( &tm, asctime_buf );
		asctime_buf[24]='\0';

		switch (record->request_method)
		{
			case HTTP_METHOD_GET:	method = "GET";		break;
			case HTTP_METHOD_HEAD:	method = "HEAD";	break;
			case HTTP_METHOD_POST:	method = "POST";	break;
		}

		switch ( record->priority )
		{
			case LOG_INFO:
				log_len = snprintf( log_msg, LOG_RECORD_LEN, "%15s [%s] \"%4s %s\" %d %s %d\n", inet_ntoa(record->client_addr), asctime_buf, method, record->request_url, record->response_status, record->msg, record->response_content_length );
				fwrite( log_msg, log_len, 1, logfile->log_access );
			break;
			default:
				log_len = snprintf( log_msg, LOG_RECORD_LEN, "[%s] [error] [client %15s] %s: %s, referer: %s\n", asctime_buf, inet_ntoa(record->client_addr), record->msg, record->request_url, record->request_referer );
				fwrite( log_msg, log_len, 1, logfile->log_error );
			break;
		}

		free( record );
	}

	//fflush( logfile->log_access );
	//fflush( logfile->log_error );
	return;
}

short tlog_open_logfiles( struct tm *now )
{
	int	i;

	for ( i=tlog.logfile_len; i--; )
	{
		pthread_mutex_init( &(tlog.logfile[i].mutex), NULL );
		if ( tlog_open_logfile( &(tlog.logfile[i]), now ) != TRUE )
		{
			tlog_close_logfile( &(tlog.logfile[i]) );
			return FALSE;
		}
	}

	return TRUE;
}

short tlog_open_logfile( struct logfile_t *logfile, struct tm *now )
{
	char	filename[LOG_FILENAME_LEN];

	/* access log open */
	if ( snprintf( filename, LOG_FILENAME_LEN, logfile->log_filename_access, (now->tm_year+1900), (now->tm_mon+1), now->tm_mday, now->tm_hour ) < 0 )
		return FALSE;
	logfile->log_access = fopen( filename, "a" );
	if ( logfile->log_access == NULL )
		return FALSE;

	logfile->log_open_flag = logfile->log_open_flag | LOG_OPEN_FLAG_ACCESS;

	/* error log open */
	if ( snprintf( filename, LOG_FILENAME_LEN, logfile->log_filename_error, (now->tm_year+1900), (now->tm_mon+1), now->tm_mday, now->tm_hour ) < 0 )
		return FALSE;

	logfile->log_error = fopen( filename, "a" );
	if ( logfile->log_access == NULL )
		return FALSE;

	logfile->log_open_flag = logfile->log_open_flag | LOG_OPEN_FLAG_ERROR;

	logfile->time = mktime( now );

	return TRUE;
}

void tlog_close_logfile( struct logfile_t *logfile )
{
	if ( logfile->log_open_flag & LOG_OPEN_FLAG_ACCESS )
		fclose( logfile->log_access );
	if ( logfile->log_open_flag & LOG_OPEN_FLAG_ERROR )
		fclose( logfile->log_error );

	logfile->log_open_flag = LOG_OPEN_FLAG_NONE;

	return;
}
